#ifndef _CARDSTMP_H
#define _CARDSTMP_H



#include "..\types\types.h"         // typedef Bit32, class Bit64
#include "..\intel\timestmp.h"      // class PentiumClock



#define CARD_CLOCK_RATE_INITIAL     25E6
#define CARD_CLOCK_RATE             CardClock::clock_rate



// 64-bit value containing a count of i960 CPU clocks
// since the FORE PCA200E card timestamp was last cleared
//
class CardClock : public Bit64
{

public:
    // latest estimate of card clock rate
    // assuming that host clock doesn't drift
    //
    static double clock_rate;

    // what was host absolute time when
    // card clock counts were cleared?
    //
    static PentiumClock sync_time;

    // no destructor needed

    // construct w/o args
    //
    CardClock()
    {
    }

#if 1
    // construct from a Bit64
    //
    CardClock( const Bit64 &second )
    :
        Bit64( second )
    {
    }
#endif

    // construct specifically from an absolute PentiumClock
    // so we don't end up doing it through our Bit64,
    // constructor, which would interpret the high
    // and low fields as card clock counts
    //
    CardClock( const PentiumClock &second )
    {
        PentiumClock dummy = second - sync_time;
        put_seconds( dummy.get_seconds() );
    }

    // construct from double-precision floating
    // point count of clocks using that assignment
    // operator
    //
    CardClock( double second )
    :
        Bit64( second )
    {
    }

    // remove bitfields inserted by the card
    // so we can operate mathematically as a Bit64 datatype
    //
    // assumes those bitfields are in the low 16 bits
    // of low 32-bit word
    //
    // also assumes that right-shifting will clear the upper bits
    //
    // also assumes that Bit32 datatype only has 32 bits,
    // by assuming that left-shifting it 16 bits will get
    // rid of the original top 16 bits
    //
    Bit16 strip_bitfields( void )
    {
        Bit16 result = low & 0xFFFF;
        low = (low >> 16) | (high << 16);
        high = high >> 16;
        return result;
    }

    // assignment from seconds
    //
    void put_seconds( const double second )
    {
        *this = second * CARD_CLOCK_RATE;
    }

    // conversion to seconds
    //
    double get_seconds() const
    {
        return ((double) *this) / CARD_CLOCK_RATE;
    }

    // use this CardClock and current CPU time
    // to get another estimate of card clock rate
    //
    // assumes that strip_bitfields() has already been called
    //
    // alters the caller's current CPU time variable because
    // (1) we cannot use stack variables at interrupt time
    // which includes implicit temporaries
    // and (2) it is more efficient than having a class
    // static variable just for this function to use as
    // a temporary and (3) the caller is probably just
    // giving us a temporary variable anyways
    //
    void recalibrate_clock_rate( PentiumClock &now ) const
    {
        now -= sync_time;
        clock_rate = double( *this ) / now.get_seconds();
    }

#if 1
    // automatic conversion to absolute PentiumClock
    //
    // if we don't do this, when a user assigns us to
    // a PentiumClock, the compiler will convert to the
    // base class Bit64 we share with PentiumClock, which
    // means the high and low fields will not be altered!
    //
    // ERROR!  this doesn't work; if you cast explicitly
    // the compiler can't decide whether to use this or
    // to treat it as Bit64 and construct PentiumCLock
    // from that; if you try a simple assignment, the
    // compiler totally barfs
    //
    // all that works is that it at least prevents implicit
    // assignment
    //
    // seems like the compiler would only really be happy
    // with a constructor inside class PentiumClock, but
    // I think that's putting the knowledge in the wrong
    // place, since
    //
    operator PentiumClock() const
    {
        PentiumClock result;

        result.put_seconds( get_seconds() );
        result += sync_time;

        return result;
    }
#endif
};



#if 0
// automatic conversion to absolute PentiumClock
//
// like the above, but done as an auxilliary
// function, not inside either class
//
// ERROR!  this doesn't work
//
operator PentiumClock( const CardClock &second )
{
    PentiumClock result;

    result.put_seconds( second.get_seconds() );
    result += CardClock::sync_time;

    return result;
}
#endif



#endif          // not included yet
