#include <stdio.h>                      // printf()
#include <stdlib.h>                     // exit()
#include <string.h>                     // strlwr()
#include "..\dpmi32\phys.h"             // map_physical() PHYS_RAM_START PHYS_RAM_END
#include "..\integrat\flat.h"



// manifest constants
//
#define MEM_USAGE_ENVAR     "DPMI32"
#define MEM_USAGE_ENVAR_2   "DPMIMEM"
#define MEM_USAGE_SCALE     1024



// global variables
//
Bit8 *flat_pool;
Bit8 *flat_free;
Bit32 flat_start;
Bit32 flat_length;
Bit32 flat_dpmi_handle;



// someday this might ask XMS to allocate a block
// of high memory and give us the physical address
// instead of assuming where DPMI32 started and
// assuming how much memory is in the machine
//
// if we do that, we better remember the XMS handle
//
void flat_init( int is_verbose_wanted )
{
    char *mem_usage_string;
    unsigned mem_usage_integer;
    int was_first_envar_used = TRUE;
    int result;

    // get the same environment variable that Borland DPMI32VM.OVL
    // used to limit its memory usage
    //
    mem_usage_string = getenv( MEM_USAGE_ENVAR );
    if( mem_usage_string == NULL )
    {
        // try the other environment variable
        //
        was_first_envar_used = FALSE;
        mem_usage_string = getenv( MEM_USAGE_ENVAR_2 );
        if( mem_usage_string == NULL )
        {
            fprintf( stderr, "flat_init() error: neither environment variable '%s' not '%s' is set\n"
                "so all of memory is probably in use, so forget it!\n",
                MEM_USAGE_ENVAR, MEM_USAGE_ENVAR_2 );
            exit( 1 );
        }
    }

    // convert the string to lowercase
    //
    strlwr( mem_usage_string );

    // make sure the keyword we want is at the beginning
    // of the environment variable
    //
    result = sscanf( mem_usage_string, " maxmem %u", &mem_usage_integer );
    if( result != 1 )
    {
        fprintf( stderr, "flat_init() error: environment variable '%s' value '%s' does not have expected format\n",
            was_first_envar_used ? MEM_USAGE_ENVAR : MEM_USAGE_ENVAR_2,
            mem_usage_string
        );

        exit( 2 );
    }

    // multiply the environment variable value
    // by its scale factor to get a byte count
    //
    mem_usage_integer *= MEM_USAGE_SCALE;

    // now calculate where it left off in memory, and how
    // much is still available for us to manage ourselves
    //
    flat_start = PHYS_RAM_START + mem_usage_integer;
    flat_length = PHYS_RAM_END - flat_start;

    if( is_verbose_wanted )

        // blab about what we decided to do
        //
        fprintf( stderr, "Borland DPMI32 server was told to use %u bytes via environment variable\n"
            "\t'%s'='%s', and we assume that previous allocations\n"
            "\tleft off at 0x%.8X physical, and we assume that there are %u\n"
            "\tbytes of RAM total on this machine, so we will use %u bytes\n"
            "\tstarting at 0x%.8X physical\n",
            mem_usage_integer,
            was_first_envar_used ? MEM_USAGE_ENVAR : MEM_USAGE_ENVAR_2,
            mem_usage_string,
            PHYS_RAM_START,
            PHYS_RAM_END,
            flat_length,
            flat_start
        );

    // map 53M of physical memory from D10000 to top of memory into our
    // process's address space so that it's addressable via DS
    //
    // or when machine has 128M instead of 64MB total, map 117MB
    //
    flat_free = flat_pool = (Bit8 *) map_physical( flat_start, flat_length, &flat_dpmi_handle );
}



void flat_exit( void )
{
    // cleanup for physical memory access
    //
    unmap_physical( flat_length, flat_dpmi_handle );
}



// allocate the given number of bytes from physical memory
//
// does not maintain any control structures, so it is
// not possible to free flat memory, let alone do so
// out of order!
//
//
//
// we are not using malloc to allocate this memory,
// because I cannot find documentation on how to
// convert a linear address to a physical address so
// it can be passed to a network card for DMA
//
// DPMI int 31h function AX=0600h locks a linear region
//
// if and when we ever get the above resolved, remember
// that pages will not necessarily be contiguous in
// physical address space, so BufferBlock could only
// have 85+1/3 48-byte buffers in it with a 4K page
// size because it only has one pointer to physical memory
// (or we could rearchitect BufferBlock to have one
// pointer per buffer, and just memcpy them all at
// once in atm_populate_ring() )
//
// but since the DMA built into the PC motherboard, which
// we eventually wanted to use for writing to disk, does
// not support scatter-gather, it is much easier to do
// as single DMA to write lots of data to to disk if it
// is stored contiguously in physical memory
//
void *flat_alloc( Bit32 length )
{
    if( flat_free + length > flat_pool + flat_length )
    {
        printf( "fatal internal error: out of locked RAM!\n" );
        exit( 0 );
    }

    void *pointer = flat_free;
    flat_free += length;
    return pointer;
}



