// similar to traceprt.cpp except:
//
// this program assumes that at all cells were captured,
// that ATM header is present for each one, and that
// that payload may not have been captured
//
// card number is a mandatory command-line argument
// to this program, since it does not use cardnum as
// an index bit in its VPI/VCI state table
//
// assumes that only the low 16 bits of timestamp are
// significant since timer interrupt handler on card
// was disabled
//
// assumes that no more than 65535 card timer ticks
// (2.621 mS == 926 cell times with 25MHz card clock)
// elapse between one cell and another in the same trace
// (any VPI/VCI)
//



// force these variables so this executable will always
// operate on an ATM header trace, regardless how CAPTURE.EXE
// is compiled (as determined by CONSTANT.H)
//
// this is only safe to do if this is the only source file
// in this executable which cares about tracerec structure
//
#define CAPTURE_TO_DISK             FALSE
#define CELLS_PER_BUFFER            0

#include <mem.h>                        // memset()
#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "trace.h"
#include "..\intel\byteswap.h"          // htons() htonl()
#include "flat.h"                       // flat_init() flat_exit() flat_alloc()



// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define PRINT_TIMES_WITH_NO_PACKETS     FALSE
#define PRINT_LENGTHS_WITH_NO_PACKETS   FALSE
#define TRACE_ALREADY_SWAPPED           TRUE
#define PRINT_MIDDLE_CELLS              FALSE



// manifest constants
//
#define NUM_VPVC                16777216L   // using all 16 VCI and all 8 UNI VPI
#define NUM_CELLS_PER_PDU       256         // allows up to 12040 bytes payload
#define NUM_DELTA_TIME_BINS     32



// data structures
//
typedef struct
{
    Bit32 start_time;           // 171 seconds with 25 MHz card clock
    Bit8 cell_count;
} ReassemblyState;



// global variables
//
struct traceblock buffer;
ReassemblyState *all_states;
Bit32 pdu_length_usage[ NUM_CELLS_PER_PDU ];    // measured in cells
Bit32 states_used = 0;
Bit32 max_states_used = 0;
Bit32 delta_time_usage[ NUM_DELTA_TIME_BINS ];


Bit8 log2( Bit32 value )
{
    Bit8 count;

    for( count=32; count>0; count-- )
        if( (1L << (count - 1)) & value )
            return count - 1;

    return 0;
}



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
    const char *outfile = argv[ 2 ];
    const char *cardstr = argv[ 3 ];
    FILE *in_handle;
    FILE *out_handle;
    int is_verbose_mode;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 prev_count;
    Bit32 block_count = 0;
    Bit32 valid_packets = 0;
    Bit32 valid_cells = 0;
    Bit32 wrong_cardnum_cells = 0;
    struct tracerec *cell;
    Bit16 cardnum_wanted;
    Bit16 when_high = 0;
    Bit16 when_low, prev_when_low;
    Bit32 when;
    Bit32 delta_time;
    Bit32 max_delta_time = 0;
    Bit8 max_delta_time_cell_count;
    AtmHeaderCoerce header;
    ReassemblyState *state;
    Bit8 cell_count;

    fprintf( stdout, "each record is %d bytes\n", sizeof(struct tracerec) );
    fprintf( stdout, "%d records in a block\n", COMPL_RING_ENTRIES );
    fprintf( stdout, "%d bytes per reassembly state table entry\n", sizeof(ReassemblyState) );
    fprintf( stdout, "%d state table entries required\n\n", NUM_VPVC );

    if( argc < 4 )
    {
        fprintf( stderr, "extracts the maximum interleave from an OC3MON ATM header trace file\n" );
        fprintf( stderr, "usage: %s infile outfile hex-cardnum {verbose}\n", program );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

    out_handle = fopen( outfile, "w" );
    if( out_handle == NULL )
    {
        fprintf( stderr, "error %d opening output file\n", errno );
        return 3;
    }

    result = sscanf( cardstr, "%hX", &cardnum_wanted );
    if( result != 1 )
    {
        fprintf( stderr, "unable to convert cardnum '%s' to integer\n", cardstr );
        return 4;
    }

    // any string in last argument slot causes verbose mode
    //
    is_verbose_mode = (argc == 5);

    // initialize flat memory management
    //
    // this is just so we can get access to more than
    // the 64 MB that Borland's DOS extender can give
    // with malloc()
    //
    flat_init();

    // try to allocate the state table
    //
    all_states = (ReassemblyState *) flat_alloc( sizeof(ReassemblyState) * NUM_VPVC );
    if( !all_states )
    {
        fprintf( stderr, "error allocating reassembly state table!\n" );
        return 5;
    }

    // initialize state table entries for all VPI and all VCI
    //
    memset( all_states, 0, sizeof(ReassemblyState)*NUM_VPVC );

    // initialize packet length histogram to empty
    //
    memset( pdu_length_usage, 0, sizeof(delta_time_usage) );

    // initialize delta time histogram to empty
    //
    memset( delta_time_usage, 0, sizeof(delta_time_usage) );

    // print header
    //
    if( is_verbose_mode )
    {
        fprintf( out_handle, "beg\t   \t     \t          \t     \t          \t\n" );
        fprintf( out_handle, " or\t   \t     \t       abs\tpending\t     delta\t# of\n" );
        fprintf( out_handle, "end\tVPI\t  VCI\t      time\treassem\t      time\tcells\n" );
        fprintf( out_handle, "---------------------------------------------------------------------\n" );
    }

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        block_count++;
        fprintf( stderr, "%lu ", block_count*sizeof(struct traceblock) );

        result = fread( &buffer, sizeof(buffer), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file\n", errno );

            break;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the usage count and card number
        //
        buffer.used = htonl( buffer.used );
        buffer.cardnum = htons( buffer.cardnum );
#endif

        // must remember usage count
        //
        // at this point it is in host byte order
        //
        used = buffer.used;

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            fprintf( stderr, "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        if( cardnum_wanted!=0 && buffer.cardnum!=cardnum_wanted )
        {
            fprintf( stderr, "ignoring cardnum 0x%.4X\n", buffer.cardnum );
            wrong_cardnum_cells += used;
            continue;
        }

        // for each record in the buffer...
        //
        for( count=0; count<used; count++ )
        {
            // shortcut for code that follows
            //
            cell = &buffer.data[ count ];

            valid_cells++;

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
            // swap the bytes and words of the timestamp
            // so we can understand them (they are in
            // network byte order on disk)
            //
            when_low = (ntohl( cell->when.high ) >> 16) & 0xFFFF;
#else
            // just grab bits 16 to 31
            //
            when_low = (cell->when.low >> 16) & 0xFFFF;
#endif

            // deal with timer wrap
            //
            if( when_low < prev_when_low )
                when_high++;
            prev_when_low = when_low;

            // construct the 32-bit global timestamp
            // from 16-bit local hardware timestamp
            // and our 16-bit software expansion
            //
            when = (when_high << 16) | when_low;

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
            header.whole = ntohl( cell->atm.whole );
#else
            header = cell->atm;
#endif

            // point at the state for this VPI/VCI
            //
            // see ATM.H for description of ATM header
            //
            state = all_states + ((header.whole >> 4) & (NUM_VPVC - 1));

            // if this is the first cell for this VPI/VCI...
            //
            if( state->cell_count == 0 )
            {
                // remember when the first cell occurred
                //
                state->start_time = when;

                // keep track of the current and max active VP/VC entries
                //
                states_used++;
                if( max_states_used < states_used )
                    max_states_used = states_used;

                if( is_verbose_mode )
                {
                    fprintf( out_handle, "beg\t%3d\t%5d\t%10lu\t%5d\n", header.bits.VPI, VCI(header.bits), when, states_used );
                }
            }
#if PRINT_MIDDLE_CELLS
            else
                if( is_verbose_mode )
                {
                    fprintf( out_handle, "mid\t%3d\t%5d\t%10lu\t%5d\n", header.bits.VPI, VCI(header.bits), when, states_used );
                }
#endif

            // if this is the last cell in the PDU...
            //
            if( header.bits.LAST )
            {
                valid_packets++;

                cell_count = state->cell_count + 1;

                // update the count of PDU's with this length
                //
                pdu_length_usage[ cell_count ]++;

                // keep track of delta time and its max
                //
                delta_time = when - state->start_time;
                if( max_delta_time < delta_time )
                {
                    max_delta_time = delta_time;
                    max_delta_time_cell_count = state->cell_count + 1;
                }
                delta_time_usage[ log2( delta_time ) ]++;

                // keep track of the number of VP/VC state entries in use
                //
                states_used--;

                // forget we have seen cells on this VPI/VCI
                //
                state->cell_count = 0;

                if( is_verbose_mode )
                {
                    fprintf( out_handle, "end\t%3d\t%5d\t%10lu\t%5d\t%10lu\t%3d\n", header.bits.VPI, VCI(header.bits), when, states_used, delta_time, cell_count );
                }
            }
            else
                state->cell_count++;
        }
    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    // show statistics
    //
    fprintf( stdout, "%lu blocks examined\n", block_count );
    fprintf( stdout, "%lu cells without cardnum 0x%.4X found\n", wrong_cardnum_cells, cardnum_wanted );
    fprintf( stdout, "%lu valid cells found\n", valid_cells );
    fprintf( stdout, "%lu valid packets found\n", valid_packets );
    fprintf( stdout, "%lu max simultaneous reassemblies\n", max_states_used );
    fprintf( stdout, "%lu outstanding reassemblies on exit\n", states_used );
    fprintf( stdout, "%lu ticks (%g seconds) max delta time, with %lu cells (%lu bytes)\n",
        max_delta_time, max_delta_time/CARD_CLOCK_RATE,
        max_delta_time_cell_count, CELL_SIZE*max_delta_time_cell_count
    );

    fprintf( stdout, "number (in decimal) of PDUs for each delta time (in decimal):\n" );
    prev_count=0;
    for( count=0; count<NUM_DELTA_TIME_BINS; count++ )    // assumes count has >16 bits
    {
#if !PRINT_TIMES_WITH_NO_PACKETS
        if( delta_time_usage[count] > 0 )
#endif
            fprintf( stdout, "%6u-%-6u %lu\n", prev_count, (1<<(count+1))-1, delta_time_usage[count] );
        prev_count = 1 << (count + 1);
    }

    fprintf( stdout, "\nnumber (in decimal) of PDUs for each length (in decimal):\n" );
    for( count=0; count<NUM_CELLS_PER_PDU; count++ )
    {
#if !PRINT_LENGTHS_WITH_NO_PACKETS
        if( pdu_length_usage[count] > 0 )
#endif
            fprintf( stdout, "%5u %lu\n", CELL_SIZE*count, pdu_length_usage[count] );
    }

    fclose( out_handle );
    fclose( in_handle );

    // return successfully
    //
    return 0;
}

