#include <mem.h>                        // memset()
#include <stdio.h>                      // fopen()
#include <errno.h>                      // errno
#include <conio.h>                      // kbhit()
#include "trace.h"
#include "..\intel\byteswap.h"          // htons() htonl()
#include "buffer.h"                     // BufferBlock
#include "fs2flows.h"                   // flows_init() flows_cleanup() flows_per_block()



// conditional compilation switches
//
#define ALLOW_KEYPRESS_TO_HALT          TRUE
#define TRACE_ALREADY_SWAPPED           TRUE



// global variables
//
Bit8 header[ SECTOR_SIZE ];
traceblock *buffer = (traceblock *) header;
BufferBlock *block;
PentiumClock fake_interval;
int fake_argc = 4;
char *fake_argv[] =
{
    "traceflo",
    "-k",
    "-t",
    "64"
};



int main( int argc, char *argv[] )
{
    const char *program = argv[ 0 ];
    const char *infile  = argv[ 1 ];
#if WANT_TIMES
    const char *outfile = argv[ 2 ];
    const char *cardstr = argv[ 3 ];
#else
    const char *cardstr = argv[ 2 ];
#endif
    FILE *in_handle;
    FILE *out_handle;
    int result;
    Bit32 used;
    Bit32 count;
    Bit32 block_count = 0;
    Bit32 valid_packets = 0;
    Bit32 wrong_cardnum_packets = 0;
    Bit32 when_temp;
    struct tracerec *packet;
    Bit16 cardnum_wanted;

    fprintf( stdout, "each record is %d bytes\n", sizeof(struct tracerec) );
    fprintf( stdout, "ICMP alone accounts for %d bytes\n", sizeof(struct icmp) );
    fprintf( stdout, "%d records in a block\n\n", COMPL_RING_ENTRIES );

    if( argc < 2 )
    {
        fprintf( stderr, "calculates flows for blocks in an OC3MON trace file\n" );
        fprintf( stderr, "usage: %s infile {hex-cardnum}\n", program );
        return 1;
    }

    in_handle = fopen( infile, "rb" );
    if( in_handle == NULL )
    {
        fprintf( stderr, "error %d opening input file\n", errno );
        return 2;
    }

    if( argc >= 3 )
    {
        result = sscanf( cardstr, "%hX", &cardnum_wanted );
        if( result != 1 )
        {
            fprintf( stderr, "unable to convert cardnum '%s' to integer\n", cardstr );
            return 4;
        }
    }
    else
        cardnum_wanted = 0;

    flat_init();
    block = new BufferBlock;

    CardClock::sync_time.set();

    result = flows_init( fake_argc, fake_argv, &fake_interval, &fake_interval, &fake_interval );
    if( result )
    {
        printf( "error %d initializing flows\n", result );
        return 5;
    }

    while( !feof( in_handle ) && !ferror( in_handle )
#if ALLOW_KEYPRESS_TO_HALT
        && !kbhit()
#endif
    )
    {
        block_count++;
        fprintf( stderr, "%lu ", block_count*sizeof(struct traceblock) );

        // read the next block header
        //
        result = fread( header, sizeof(header), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d header\n", errno, block_count );

            break;
        }

        // read the trace records into the BufferBlock
        //
        // note that this must NOT be conditional on our
        // liking the block, or we will be unable to read
        // the next block!
        //
        result = fread( block->buffers, COMPL_RING_ENTRIES*sizeof(tracerec), 1, in_handle );
        if( result != 1 )
        {
            if( !feof( in_handle ) )
                fprintf( stderr, "error %d reading input file block %d data\n", errno, block_count );

            break;
        }

#if HOST_IS_BIG_ENDIAN != TRACE_ALREADY_SWAPPED
        // swap the usage count and card number
        //
        buffer->used = htonl( buffer->used );
        buffer->cardnum = htons( buffer->cardnum );
#endif

        // must remember usage count
        //
        // at this point it is in host byte order
        //
        used = buffer->used;

        // check for out-of-bound record count
        //
        if( used > COMPL_RING_ENTRIES )
        {
            fprintf( stderr, "error: usage count of block %u is %u, but max is %u!\n", block_count, used, COMPL_RING_ENTRIES );
            break;
        }

        if( cardnum_wanted!=0 && buffer->cardnum!=cardnum_wanted )
        {
            fprintf( stderr, "ignoring cardnum 0x%.4X\n", buffer->cardnum );
            wrong_cardnum_packets += used;
            continue;
        }

        valid_packets += used;

        // copy the fields we like to the BufferBlock
        //
        block->cardnum = buffer->cardnum;
        block->cardindex = buffer->cardindex;
        block->used = buffer->used;
        block->queried = 0;

        // for all records in the BufferBlock
        //
        for( count=0; count<block->used; count++ )
        {
            packet = (tracerec *) (block->buffers + count*BUFFER_SIZE);

            // swap the bytes and words of the timestamp
            //
            when_temp = htonl( packet->when.low );
            packet->when.low  = htonl( packet->when.high );
            packet->when.high = when_temp;

            // swap the bytes of the ATM header
            //
            packet->atm.whole = htonl( packet->atm.whole );

            // the rest of the data is IP and TCP headers,
            // and was always in network byte order!
            //
        }

        // hand it off to flows routines
        //
        result = flows_per_block( block );
        if( result )
        {
            fprintf( stderr, "error %d processing flows in block %d\n", result, block_count );
            break;
        }
    }
    fprintf( stderr, "\n" );

#if ALLOW_KEYPRESS_TO_HALT
    if( kbhit() )
    {
        getch();
        fprintf( stderr, "halted by keypress!\n" );
    }
#endif

    flows_cleanup();
    delete block;
    flat_exit();

    // show statistics
    //
    fprintf( stdout, "%lu blocks examined\n", block_count );
    if( cardnum_wanted != 0 )
        fprintf( stdout, "%lu packets without cardnum 0x%.4X found\n", wrong_cardnum_packets, cardnum_wanted );
    fprintf( stdout, "%lu valid packets found\n", valid_packets );

    fclose( in_handle );

    // return successfully
    //
    return 0;
}

